import datetime
import json
import unittest

import jsonschema

from .. import types


class TestString(unittest.TestCase):

    def test_schema(self):
        t = types.String()

        self.assertEqual(t.get_schema(), {"type": "string"})

    def test_validation_pattern(self):
        t = types.String(pattern=r".+")

        t.validate("valid")
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate("")


class TestEnum(unittest.TestCase):

    def test_schema(self):
        t = types.Enum()

        self.assertEqual(t.get_schema(), {})

    def test_validation_enum(self):
        t = types.Enum(enum=["a", "b", 1, True])
        self.assertEqual(t.get_schema(), {"enum": ["a", "b", 1, True]})

        t.validate("a")
        t.validate(1)
        t.validate(True)
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate("c")
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate(2)
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate(False)


class TestInteger(unittest.TestCase):

    def test_schama(self):
        t = types.Integer()

        self.assertEqual(t.get_schema(), {"type": "integer"})

    def test_validation(self):
        t = types.Integer()

        t.validate(12)
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate(12.22)

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate("Hello")


class TestNumber(unittest.TestCase):

    def test_schema(self):
        t = types.Number()

        self.assertEqual(t.get_schema(), {"type": "number"})

    def test_validation(self):
        t = types.Number()

        t.validate(12)
        t.validate(12.22)

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate("Hello")


class TestBoolean(unittest.TestCase):

    def test_schema(self):
        t = types.Boolean()

        self.assertEqual(t.get_schema(), {"type": "boolean"})

    def test_validation(self):
        t = types.Boolean()

        t.validate(True)
        t.validate(False)

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate("Hello")


class TestArray(unittest.TestCase):

    def test_schema(self):
        t = types.Array()

        self.assertEqual(t.get_schema(), {"type": "array"})

    def test_validation(self):
        t = types.Array()
        t.validate([])
        t.validate([1, 2, 3])

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate("Hello")

    def test_validation_unique_items(self):
        t = types.Array(unique_items=True)
        t.validate([])
        t.validate([1, 2, 3])

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate([1, 2, 2])

    def test_validation_min_max_items(self):
        t = types.Array(min_items=2, max_items=4)
        t.validate([1, 2, 3])

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate("Hello")

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate([])

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate([1])

        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate([1, 2, 3, 4, 5])

    def test_schema_with_items(self):
        t = types.Array(items=types.String(), additional=False)

        self.assertEqual(
            t.get_schema(),
            {
                'type': 'array',
                'items': {'type': 'string'},
                'additionalItems': False
            }
        )
        t.validate(['1', '2', '3'])
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate([1, 2, 3])

    def test_schema_with_items_list(self):
        t = types.Array(items=[types.String()])

        self.assertEqual(
            t.get_schema(),
            {
                'type': 'array',
                'items': [{'type': 'string'}],
            }
        )
        t.validate(['1', '2', '3'])
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate([1, 2, 3])

    def test_schema_with_items_list_strict(self):
        t = types.Array(items=[types.String()], additional=False)

        self.assertEqual(
            t.get_schema(),
            {
                'type': 'array',
                'items': [{'type': 'string'}],
                'additionalItems': False
            }
        )
        t.validate(['1'])
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate(['1', '2', '3'])
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate([1, 2, 3])


class TestDate(unittest.TestCase):

    def test_validation(self):
        t = types.Date()

        t.validate(datetime.date(2012, 12, 24))
        t.validate("2012-12-24")
        with self.assertRaises(ValueError):
            t.validate('crappy')

    def test_serialize(self):
        t = types.Date()
        v = t.dump(datetime.date(2012, 12, 24))
        self.assertEqual(v, '"2012-12-24"')

    def test_loads(self):
        t = types.Date()

        v = t.load('"2012-12-24"')
        self.assertEqual(v, datetime.date(2012, 12, 24))

    def test_complex(self):
        class MySchema(types.Object):
            date = types.Date()
        t = MySchema()
        t.validate({"date": datetime.date(2012, 12, 24)})
        t.dump({"date": datetime.date(2012, 12, 24)})


class TestDuration(unittest.TestCase):

    def test_validation(self):
        t = types.Duration()

        t.validate(datetime.timedelta(seconds=12))
        t.validate(12)
        with self.assertRaises(ValueError):
            t.validate('crappy')

    def test_serialize(self):
        t = types.Duration()
        v = t.dump(datetime.timedelta(seconds=12))
        self.assertEqual(v, '"PT12S"')

    def test_loads(self):
        t = types.Duration()

        v = t.load('"PT12S"')
        self.assertEqual(v, datetime.timedelta(seconds=12))

        v = t.load('12')
        self.assertEqual(v, datetime.timedelta(seconds=12))

        v = t.load('12.1')
        self.assertEqual(v, datetime.timedelta(seconds=12, milliseconds=100))

    def test_complex(self):
        class MySchema(types.Object):
            duration = types.Duration()
        t = MySchema()
        t.validate({"duration": datetime.timedelta(seconds=12)})
        t.dump({"duration": datetime.timedelta(seconds=12)})


class TestObject(unittest.TestCase):

    def test_serialize(self):
        class MySchema(types.Object):
            username = types.String()
            password = types.String()

        data = {'username': 'admin', 'password': 'secret', 'pk': 1}
        dumped = MySchema(additional=None).dump(data)
        self.assertEqual(
            json.loads(dumped),
            {'username': 'admin', 'password': 'secret', 'pk': 1}
        )
        self.assertEqual(
            data,
            {'username': 'admin', 'password': 'secret', 'pk': 1}
        )



import unittest

import jsonschema

from .. import types


class TestSchemaValidation(unittest.TestCase):

    def test_declarative_schema(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(name="String")

        t = MyObject(additional=None)

        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "num": {"type": "integer"},
                    "String": {"type": "string"},
                }
            }
        )

    def test_description_as_docstring(self):
        class MyObject(types.Object):
            """Description"""
            num = types.Integer()

        t = MyObject(additional=None)

        self.assertEqual(
            t.get_schema(),
            {
                'type': 'object',
                'description': "Description",
                'properties': {'num': {'type': 'integer'}}
            }
        )

    def test_declarative_schema_required(self):
        class MyObject(types.Object):
            num = types.Integer(required=True)
            string = types.String(name="String", required=True)

        t = MyObject(additional=None)
        s = t.get_schema()
        expected = {
            "type": "object",
            "properties": {
                "num": {"type": "integer"},
                "String": {"type": "string"},
            },
            "required": ['String', 'num']
        }
        self.assertEqual(
            s, expected
        )

    def test_declarative_schema_validation(self):
        class MyObject(types.Object):
            string = types.String(name="String")

        t = MyObject()
        t.validate({"string": "value"})

    def test_subschema(self):
        class MySubObject(types.Object):
            integer = types.Integer(name="int")

        class MyObject(types.Object):
            string = types.String(name="String")
            sub = MySubObject(null=True, additional=None)
        t = MyObject(additional=None)

        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "String": {"type": "string"},
                    "sub": {
                        "type": ["object", "null"],
                        "properties": {"int": {"type": "integer"}}
                    },
                }
            }
        )
        t.validate({"string": "hello", "sub": {"int": 12}})
        t.validate({"string": "hello", "sub": None})

    def test_definitions(self):
        class MySubObject(types.Object):
            integer = types.Integer(name="int")

        class MyObject(types.Object):
            string = types.String(name="String")
            sub = types.Ref(ref="sub")

            class Definitions:
                sub = MySubObject(null=True, additional=None)
        t = MyObject(additional=None)

        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "String": {"type": "string"},
                    "sub": {"$ref": "#/definitions/sub"}
                },
                "definitions": {
                    "sub": {
                        "type": ["object", "null"],
                        "properties": {"int": {"type": "integer"}}
                    }
                }
            }
        )
        t.validate({"String": "hello", "sub": {"int": 12}})
        t.validate({"String": "hello", "sub": None})

    def test_deep_definitions(self):
        class MySubObject(types.Object):
            integer = types.Ref(ref="integer")

            class Definitions:
                integer = types.Integer()

        class MyObject(types.Object):
            string = types.String(name="String")
            sub = types.Ref(ref="sub")

            class Definitions:
                sub = MySubObject(null=True, additional=None)
        t = MyObject(additional=None)

        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "String": {"type": "string"},
                    "sub": {"$ref": "#/definitions/sub"}
                },
                "definitions": {
                    "sub": {
                        "type": ["object", "null"],
                        "properties": {
                            "integer": {"$ref": "#/definitions/integer"}
                        },
                        "definitions": {"integer": {"type": "integer"}}
                    }
                }
            }
        )
        t.validate({"String": "hello", "sub": {"int": 12}})
        t.validate({"String": "hello", "sub": None})


class TestSchemaToPython(unittest.TestCase):

    def test_schema_to_python(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(name="NewName")

        t = MyObject()
        p = t.to_python({"num": 1, "NewName": "hi", 'x': 'y'})
        self.assertEqual(p, {"num": 1, "string": "hi", 'x': 'y'})

    def test_special_type(self):

        class Spec(types.String):
            def to_python(self, src, context=None):
                if src.lower() == "yes":
                    return True
                else:
                    return False

        class MyObject(types.Object):
            username = types.String()
            can_login = Spec()

        t = MyObject()
        p = t.to_python({'username': 'user', 'can_login': 'Yes'})
        self.assertEqual(p, {'username': 'user', 'can_login': True})


class TestSchemaToJson(unittest.TestCase):

    def test_schema_to_json(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(name="NewName")

        t = MyObject()
        p = t.to_json({"num": 1, "string": "hi", "unknown": "x"})
        self.assertEqual(p, {"num": 1, "NewName": "hi", 'unknown': 'x'})

    def test_special_type(self):

        class Spec(types.String):
            def to_json(self, src, context=None):
                return "*******"

        class MyObject(types.Object):
            string = types.String()
            password = Spec()

        t = MyObject()
        p = t.to_json({'username': 'user', 'password': 'secret'})
        self.assertEqual(p, {'username': 'user', 'password': '*******'})


class TestSchemaAdditional(unittest.TestCase):

    def test_schema(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(name="String")

            class Attrs:
                additional = False

        t = MyObject()

        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "num": {"type": "integer"},
                    "String": {"type": "string"},
                },
                "additionalProperties": False
            }
        )

    def test_validation(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(name="String")

            class Attrs:
                additional = False

        t = MyObject()
        t.validate({'num': 12})
        t.validate({'string': 'string'})
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate({'string': 'string', 'extra': 'invalid'})


class TestSchemaPattern(unittest.TestCase):

    def test_schema(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(name="String")

            class Attrs:
                additional = False
                min_properties = 2
                max_properties = 4
                patterns = {
                    '[a-z]{2}': types.String()
                }

        t = MyObject()

        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "num": {"type": "integer"},
                    "String": {"type": "string"},
                },
                "additionalProperties": False,
                'minProperties': 2,
                'maxProperties': 4,
                'patternProperties': {
                    '[a-z]{2}': {'type': 'string'}
                }
            }
        )

    def test_validation(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(name="String")

            class Attrs:
                additional = False
                patterns = {
                    'name_[a-z]{2}': types.String()
                }

        t = MyObject()
        t.validate({'num': 12})
        t.validate({'string': 'string'})
        t.validate({'string': 'string', 'name_en': 'English'})
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            t.validate({'string': 'string', 'extra': 'invalid'})


class TestSchemaLoadForm(unittest.TestCase):

    def test_load_form(self):
        class MySub(types.Object):
            sub = types.Integer()

        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(name="NewName")
            arr = types.Array()
            sub = MySub(additional=None)

        t = MyObject(additional=None)
        form = {
            'num': '1',
            'NewName': 'hi',
            'sub': '{"sub": 1}',
            'arr': '[1, 2, "hi"]',
            'unknown': '{"any": "value"}'
        }
        r = t.load(form)
        self.assertEqual(r, {
            "num": 1,
            "string": "hi",
            'sub': {'sub': 1},
            'arr': [1, 2, 'hi'],
            'unknown': '{"any": "value"}'
        })


class TestSchemaExtend(unittest.TestCase):

    def test_procedural_extend(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String()

            class Attrs:
                additional = False

        t = MyObject()
        t.get_schema()
        t.validate({'num': 12})
        t.extend({
            'title': types.String()
        })

        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "num": {"type": "integer"},
                    "string": {"type": "string"},
                    "title": {"type": "string"},
                },
                'additionalProperties': False,
            }
        )
        t.validate({'num': 12, 'title': 'test'})

    def test_initial_extend(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String()

            class Attrs:
                additional = False

        t = MyObject(extend={'title': types.String()})
        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "num": {"type": "integer"},
                    "string": {"type": "string"},
                    "title": {"type": "string"},
                },
                'additionalProperties': False,
            }
        )
        t.validate({'num': 12, 'title': 'test'})


class TestSchemaInclude(unittest.TestCase):

    def test_partial(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String()

            class Attrs:
                additional = False

        t = MyObject(include=['num'])
        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "num": {"type": "integer"},
                },
                'additionalProperties': False,
            }
        )
        t.validate({'num': 12})


class TestSchemaExclude(unittest.TestCase):

    def test_initial_exclude(self):
        class MyObject(types.Object):
            num = types.Integer()
            string = types.String(required=True)

            class Attrs:
                additional = False

        t = MyObject(exclude=['string'])
        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "num": {"type": "integer"},
                },
                'additionalProperties': False,
            }
        )
        t.validate({'num': 12})


class TestSchemaTagFilter(unittest.TestCase):

    def test_context_exclude(self):
        class SubSchema(types.Object):
            name = types.String()
            password = types.String(tags=['sensitive'])

        class MyObject(types.Object):
            fullname = types.String()
            email = types.String(required=True, tags=['sensitive'])
            login = SubSchema(additional=None)

            class Attrs:
                additional = False

        t = MyObject()
        c = {'exclude_tags': 'sensitive'}
        self.assertEqual(
            t.get_schema(context=c),
            {
                "type": "object",
                "properties": {
                    "fullname": {"type": "string"},
                    'login': {
                        'type': 'object',
                        'properties': {
                            'name': {'type': 'string'}
                        }
                    }
                },
                'additionalProperties': False,
            }
        )
        t.validate({'fullname': 'test'})

    def test_context_exclude_initial(self):
        class SubSchema(types.Object):
            name = types.String()
            password = types.String(tags=['sensitive'])

        class MyObject(types.Object):
            fullname = types.String()
            email = types.String(required=True, tags=['sensitive'])
            login = SubSchema(additional=None)

            class Attrs:
                additional = False

        t = MyObject(exclude_tags='sensitive')
        self.assertEqual(
            t.get_schema(),
            {
                "type": "object",
                "properties": {
                    "fullname": {"type": "string"},
                    'login': {
                        'type': 'object',
                        'properties': {
                            'name': {'type': 'string'}
                        }
                    }
                },
                'additionalProperties': False,
            }
        )
        t.validate({'fullname': 'test'})

    def test_context_exclude_tags_mix(self):
        class SubSchema(types.Object):
            name = types.String()
            password = types.String(tags=['sensitive'])
            last_update = types.String(tags=['readonly'])

        class MyObject(types.Object):
            fullname = types.String()
            email = types.String(required=True, tags=['sensitive'])
            login = SubSchema(additional=None)

            class Attrs:
                additional = False

        t = MyObject(exclude_tags='sensitive')
        c = {'exclude_tags': 'readonly'}
        self.assertEqual(
            t.get_schema(context=c),
            {
                "type": "object",
                "properties": {
                    "fullname": {"type": "string"},
                    'login': {
                        'type': 'object',
                        'properties': {
                            'name': {'type': 'string'}
                        }
                    }
                },
                'additionalProperties': False,
            }
        )
        t.validate({'fullname': 'test'})



"""
This module introduce the basic schema types.
"""
import collections
import datetime
import json

import isodate
import six

from . import base
from . import formats


class String(base.Base):
    _type = "string"

    def make_schema(self, context=None):
        schema = super(String, self).make_schema(context=context)
        if self.get("pattern"):
            schema["pattern"] = self["pattern"]
        return schema

    def to_object(self, value, context=None):
        return value


class Integer(base.Base):
    _type = "integer"


class Number(base.Base):
    _type = "number"


class Boolean(base.Base):
    _type = "boolean"


class Array(base.Base):
    """
    Successful validation of an array instance with regards to these two
    keywords is determined as follows:

    if "items" is not present, or its value is an object, validation of
    the instance always succeeds, regardless of the value of "additional"

    if the value of "additional" is boolean value true or an object,
    validation of the instance always succeeds;

    if the value of "additional" is boolean value false and the value of
    "items" is an array, the instance is valid if its size is less than,
    or equal to, the size of "items".

    Array specific options:
        min_items:
            An array instance is valid against "min_items" if its size is
            greater than, or equal to, the value of this keyword.
        max_items:
            An array instance is valid against "max_items" if its size is
            less than, or equal to, the value of this keyword.
        unique_items:
            If this keyword has boolean value false, the instance validates
            successfully. If it has boolean value true, the instance validates
            successfully if all of its elements are unique.
    """
    _type = "array"

    def make_schema(self, context=None):
        schema = super(Array, self).make_schema(context=context)
        if self.get('additional') is not None:
            schema['additionalItems'] = self.get('additional')
        if self.get('max_items') is not None:
            schema['maxItems'] = self.get('max_items')
        if self.get('min_items') is not None:
            schema['minItems'] = self.get('min_items')
        if self.get('unique_items') is not None:
            schema['uniqueItems'] = self.get('unique_items')
        if self.get('items'):
            if isinstance(self.get('items'), (list, tuple)):
                schema['items'] = [
                    s.get_schema(context=context) for s in self.get('items')
                ]
            else:
                schema['items'] = self.get('items').get_schema(context=context)
        return schema


class Object(base.Base):
    """Declarative schema object

    Object specific attributes:
        additional:
            boolean value: enable or disable extra items on the object
            schema: items which are valid against the schema allowed to extend
            **false by default**
        min_properties:
            An object instance is valid against `min_properties` if its number
            of properties is greater than, or equal to, the value.
        max_properties:
            An object instance is valid against `max_properties` if its number
            of properties is less than, or equal to, the value.
        pattern:
            Should be a dict where the keys are valid regular excpressions
            and the values are schema instances. The object instance is valid
            if the extra properties (which are not listed as property) valid
            against the schema while name is match on the pattern.

        Be careful, the pattern sould be explicit as possible, if the pattern
        match on any normal property the validation should be successful
        against them as well.

        A normal object should looks like the following:

        .. code:: python

            class Translation(types.Object):
                keyword = types.String()
                value = types.String()

                class Attrs:
                    additional = False
                    patterns = {
                        'value_[a-z]{2}': types.String()
                    }
    """
    _type = "object"
    _attrs = {'additional': False}

    def __init__(self, extend=None, **attrs):
        super(Object, self).__init__(**attrs)
        if extend:
            self._fields.update(extend)

    def make_schema(self, context=None):
        schema = super(Object, self).make_schema(context=context)
        if 'description' not in schema and self.__doc__:
            schema['description'] = self.__doc__
        if self.get('additional') is not None:
            if isinstance(self.get('additional'), bool):
                schema['additionalProperties'] = self.get('additional')
            else:
                schema['additionalProperties'] = \
                    self.get('additional').get_schema(context=context)
        if self.get('min_properties') is not None:
            schema['minProperties'] = self.get('min_properties')
        if self.get('max_properties') is not None:
            schema['maxProperties'] = self.get('max_properties')
        if self.get('patterns'):
            patterns = collections.OrderedDict()
            for reg, pattern in self.get('patterns').items():
                patterns[reg] = pattern.get_schema(context=context)
            schema['patternProperties'] = patterns
        return schema

    @property
    def properties(self):
        return self._fields

    def extend(self, properties, context=None):
        """Extending the exist same with new properties.
        If you want to extending with an other schema, you should
        use the other schame `properties`
        """
        self._fields.update(properties)
        self.invalidate(context=None)


class Date(String):
    _attrs = {"format": "date"}

    def to_python(self, value, context=None):
        if isinstance(value, datetime.date):
            return value
        try:
            return isodate.parse_date(value)
        except isodate.ISO8601Error:
            raise ValueError("Invalid date '%s'" % value)

    def to_json(self, value, context=None):
        if isinstance(value, datetime.date):
            return isodate.date_isoformat(value)
        if isinstance(value, six.string_types):
            self.to_python(value)
            return value


class Time(String):
    _attrs = {"format": "time"}

    def to_python(self, value, context=None):
        if isinstance(value, datetime.time):
            return value
        try:
            return isodate.parse_time(value)
        except isodate.ISO8601Error:
            raise ValueError("Invalid time '%s'" % value)

    def to_json(self, value, context=None):
        if isinstance(value, datetime.time):
            return isodate.time_isoformat(value)
        if isinstance(value, six.string_types):
            self.to_pyton(value)
            return value


class DateTime(String):
    _attrs = {"format": "datetime"}

    def to_python(self, value, context=None):
        if isinstance(value, datetime.datetime):
            return value
        try:
            return formats.parse_datetime(value)
        except isodate.ISO8601Error:
            raise ValueError("Invalid datetime '%s'" % value)

    def to_json(self, value, context=None):
        if isinstance(value, datetime.time):
            return isodate.datetime_isoformat(value)
        if isinstance(value, six.string_types):
            self.to_pyton(value)
            return value


class Duration(String):
    _attrs = {"format": "duration"}

    def to_python(self, value, context=None):
        if isinstance(value, (int, float)):
            return datetime.timedelta(seconds=value)
        try:
            return isodate.parse_duration(value)
        except isodate.ISO8601Error:
            raise ValueError("Invalid duration '%s'" % value)

    def to_json(self, value, context=None):
        if isinstance(value, datetime.timedelta):
            return isodate.duration_isoformat(value)
        if isinstance(value, (int, float)):
            return isodate.duration_isoformat(
                datetime.timedelta(seconds=value)
            )
        if isinstance(value, six.string_types):
            self.to_python(value)
            return value


class TimeDelta(Number):

    def to_python(self, value, context=None):
        if isinstance(value, six.string_types):
            value = json.loads(value)
        if isinstance(value, (int, float)):
            return datetime.timedelta(seconds=value)
        if isinstance(value, datetime.timedelta):
            return value
        raise ValueError("Invalid type of timedelta '%s'" % type(value))

    def to_json(self, value, context=None):
        if isinstance(value, datetime.timedelta):
            return value.total_seconds()
        if isinstance(value, (int, float)):
            return value
        raise ValueError("Invalid type of timedelta '%s'" % type(value))


class Enum(base.Base):
    """JSON generic enum class

    :param enum: list of possible values
    :type enum: list
    """

    def make_schema(self, context=None):
        """Ensure the generic schema, remove `types`

        :return: Gives back the schema
        :rtype: dict
        """
        schema = super(Enum, self).make_schema(context=None)
        schema.pop("type")
        if self.get("enum"):
            schema["enum"] = self["enum"]
        return schema


class Ref(base.Base):

    def make_schema(self, context=None):
        schema = super(Ref, self).make_schema(context=context)
        schema.pop("type")
        assert not schema
        return {"$ref": "#/definitions/"+self["ref"]}



class NA:
    pass


def get_public_attributes(cls):
    if not cls:
        return {}
    return dict(
        [(k, v) for k, v in cls.__dict__.items() if not k.startswith("_")]
    )


def ensure_set(thing):
    if thing is None:
        return set()
    if isinstance(thing, set):
        return thing
    if isinstance(thing, (list, tuple)):
        return set(thing)
    return set([thing])


def ensure_list(thing):
    if thing is None:
        return list()
    if isinstance(thing, list):
        return thing
    if isinstance(thing, (set, tuple)):
        return list(thing)
    return list([thing])



import unittest

import jsonschema
import mock

from .. import base
from .. import exceptions
from .. import types


class TestBase(unittest.TestCase):

    def test_attrs(self):
        b = base.Base(attr=1)

        self.assertEqual(b.get("attr"), 1)
        self.assertEqual(list(b.keys()), ["attr"])

    def test_load(self):
        b = base.Base()
        with mock.patch.object(b, "validate_json") as validate:
            res = b.load('"Hello"')

        self.assertEqual(res, "Hello")
        validate.assert_called_with("Hello", context=None)

    def test_dump(self):
        b = base.Base()
        with mock.patch.object(b, "validate_json") as validate:
            res = b.dump("Hello")

        self.assertEqual(res, '"Hello"')
        validate.assert_called_with("Hello", context=None)

    def test_get_schema(self):
        with mock.patch.object(base.Base, "_type", new="string"):
            b = base.Base()
            s = b.get_schema()
            self.assertEqual(s, {"type": "string"})
        self.assertIsNone(base.Base._type)

    def test_validation(self):
        with mock.patch.object(base.Base, "_type", new="string"):
            b = base.Base()
            b.validate("Hello")
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                b.validate(None)
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                b.validate({"string": 1})

    def test_null_validation(self):
        with mock.patch.object(base.Base, "_type", new="string"):
            b = base.Base(null=True)
            b.validate("Hello")
            b.validate(None)

    def test_enum_validation(self):
        with mock.patch.object(base.Base, "_type", new="string"):
            b = base.Base(enum=["a", "b", "c"])

            b.validate("a")
            with self.assertRaises(jsonschema.exceptions.ValidationError):
                b.validate("Hello")

    def test_default_attrs(self):
        class MyType(base.Base):
            _type = "string"
            _attrs = {"null": True}
        t = MyType()

        self.assertEqual(t.get_schema(), {"type": ["string", "null"]})

    def test_title_description(self):
        class MyType(base.Base):
            _type = "string"
        t = MyType(title="t", description='d')

        self.assertEqual(
            t.get_schema(),
            {"type": "string", "title": "t", "description": 'd'}
        )

    def test_title_description_declarative(self):
        class MyType(base.Base):
            _type = "string"

            class Attrs:
                title = 't'
                description = 'd'
        t = MyType()

        self.assertEqual(
            t.get_schema(),
            {"type": "string", "title": "t", "description": 'd'}
        )


class TestDeclarativeBase(unittest.TestCase):

    def test_declarative_attrs_update(self):
        class MyType(base.Base):
            _type = "string"

            class Attrs:
                null = True

        t = MyType()

        self.assertEqual(t.get_schema(), {"type": ["string", "null"]})


class TestDefault(unittest.TestCase):

    def test_default_schema(self):
        class MySchema(types.Object):
            name = types.String(default='example')

        t = MySchema(additional=None)
        self.assertEqual(
            t.get_schema(),
            {
                'type': 'object',
                'properties': {
                    'name': {'type': 'string', 'default': 'example'}
                }
            }
        )


class TestSchema(unittest.TestCase):

    def test_creation_index(self):
        base.Schema._creation_index = 0

        b0 = base.Schema()
        b1 = base.Base()
        b2 = base.Base()

        self.assertEqual(b0._creation_index, 0)
        self.assertEqual(b1._creation_index, 1)
        self.assertEqual(b2._creation_index, 2)
        self.assertEqual(base.Base._creation_index, 3)

    def test_validation(self):
        class MySchema(base.Schema):
            _schema = {
                'type': 'object',
                'properties': {
                    'num': {'type': 'integer'}
                }
            }

        s = MySchema()
        s.validate({'num': 12})
        with self.assertRaises(exceptions.ValidationError):
            s.validate({'num': 12.1})

    def test_validation_inline(self):
        s = base.Schema({
            'type': 'object',
            'properties': {
                'num': {'type': 'integer'}
            }
        })
        s.validate({'num': 12})
        with self.assertRaises(exceptions.ValidationError):
            s.validate({'num': 12.1})

    def test_redeclaration_raise_error(self):
        class MySchema(base.Schema):
            _schema = {'type': 'string'}

        with self.assertRaises(AttributeError):
            MySchema({
                'type': 'object',
                'properties': {
                    'num': {'type': 'integer'}
                }
            })

    def test_load(self):
        class MySchema(base.Schema):
            _schema = {
                'type': 'object',
                'properties': {
                    'num': {'type': 'integer'}
                }
            }

        s = MySchema()
        res = s.load('{"num": 12}')
        self.assertEqual(res, {'num': 12})

    def test_dump(self):
        class MySchema(base.Schema):
            _schema = {
                'type': 'object',
                'properties': {
                    'num': {'type': 'integer'}
                }
            }

        s = MySchema()
        res = s.dump({'num': 12})
        self.assertEqual(res, '{"num": 12}')

    def test_dump_as_field(self):
        class MyType(base.Schema):
            _schema = {
                'type': 'object',
                'properties': {
                    'num': {'type': 'integer'}
                }
            }

        class MyObject(types.Object):
            code = MyType()

        s = MyObject()
        res = s.dump({'code': {'num': 12}})
        self.assertEqual(res, '{"code": {"num": 12}}')

    def test_load_as_field(self):
        class MyType(base.Schema):
            _schema = {
                'type': 'object',
                'properties': {
                    'num': {'type': 'integer'}
                }
            }

        class MyObject(types.Object):
            code = MyType()

        s = MyObject()
        res = s.load('{"code": {"num": 12}}')
        self.assertEqual(res, {'code': {'num': 12}})



import datetime
import re

import isodate
import jsonschema
import six


draft4_format_checkers = list(jsonschema.draft4_format_checker.checkers.keys())


def parse_datetime(datetimestring):
    '''
    Parses ISO 8601 date-times into datetime.datetime objects.
    This function uses parse_date and parse_time to do the job, so it allows
    more combinations of date and time representations, than the actual
    ISO 8601:2004 standard allows.
    '''
    try:
        datestring, timestring = re.split('T| ', datetimestring)
    except ValueError:
        raise isodate.ISO8601Error(
            "ISO 8601 time designator 'T' or ' '  missing. Unable to parse "
            "datetime string %r" % datetimestring
        )
    tmpdate = isodate.parse_date(datestring)
    tmptime = isodate.parse_time(timestring)
    return datetime.combine(tmpdate, tmptime)


def format_checker(name, raises=()):
    def wrap(func):
        draft4_format_checkers.append(name)
        func = jsonschema.FormatChecker.cls_checks(name, raises)(func)
        return func
    return wrap


@format_checker('date', (ValueError, isodate.ISO8601Error, ))
def date_format_checker(instance):
    if isinstance(instance, datetime.date):
        return True
    if isinstance(instance, six.string_types):
        return isodate.parse_date(instance)


@format_checker('datetime', (ValueError, isodate.ISO8601Error, ))
def datetime_format_checker(instance):
    if isinstance(instance, datetime.datetime):
        return True
    if isinstance(instance, six.string_types):
        return parse_datetime(instance)


@format_checker('time', (ValueError, isodate.ISO8601Error, ))
def time_format_checker(instance):
    if isinstance(instance, datetime.datetime):
        return True
    if isinstance(instance, six.string_types):
        return isodate.parse_time(instance)


@format_checker('duration', (ValueError, isodate.ISO8601Error, ))
def duration_format_checker(instance):
    if isinstance(instance, (datetime.timedelta, int, float)):
        return True
    if isinstance(instance, six.string_types):
        return isodate.parse_duration(instance)



import collections
import datetime
import functools
import json

import isodate
import jsonschema
import six

from . import lib
from . import formats


class Schema(object):
    _creation_index = 0
    _schema = None
    _attrs = None

    def __init__(self, _schema=None, **attrs):
        self._creation_index = Schema._creation_index
        Schema._creation_index += 1
        if _schema and self._schema:
            raise AttributeError("The declared schema shouldn't be redefined")
        if _schema:
            self._schema = _schema
        if self.__class__._attrs is not None:
            self._attrs = self.__class__._attrs.copy()
        else:
            self._attrs = collections.OrderedDict()
        self._attrs.update(attrs)

    def __getitem__(self, name):
        return self._attrs[name]

    def __setitem__(self, name, value):
        self._attrs[name] = value
        self.invalidate()

    def keys(self):
        return self._attrs.keys()

    def get(self, name, default=None):
        return self._attrs.get(name, default)

    def get_schema(self, context=None):
        return self._schema

    def get_tags(self):
        return self.get('tags', [])

    def has_tags(self, tags):
        if not isinstance(tags, (list, tuple, set)):
            tags = {tags}
        if set(self.get_tags()) & set(tags):
            return True
        return False

    def load(self, value, context=None):
        if isinstance(value, six.string_types):
            obj = json.loads(value)
            self.validate_json(obj, context=context)
            self._value = self.to_python(obj, context=context)
            return self._value
        raise ValueError('Unrecognised input format')

    def dump(self, obj, context=None):
        obj = self.to_json(obj, context=context)
        self.validate_json(obj, context=context)
        return self._dump(obj, context=context)

    def _dump(self, obj, context=None):
        default = functools.partial(self._dump_default, context=context)
        return json.dumps(obj, default=default)

    def _dump_default(self, obj, context=None):
        if isinstance(obj, datetime.datetime):
            return isodate.datetime_isoformat(obj)
        elif isinstance(obj, datetime.date):
            return isodate.date_isoformat(obj)
        elif isinstance(obj, datetime.time):
            return isodate.time_isoformat(obj)
        elif isinstance(obj, datetime.timedelta):
            return obj.total_seconds()
        else:
            raise TypeError(obj)

    def make_validator(self, context=None):
        return _make_validator(self.get_schema(context=context))

    def validate(self, obj, context=None):
        self.validate_json(self.to_json(obj, context=context))

    def validate_json(self, obj, context=None):
        self.get_validator(context=context).validate(obj)

    def get_validator(self, context=None):
        if context:
            return self.make_validator(context=context)
        if getattr(self, "_validator", None):
            return self._validator
        self._validator = self.make_validator(context=context)
        return self._validator

    def to_json(self, value, context=None):
        """Convert the value to a JSON compatible value"""
        return value

    def to_python(self, value, context=None):
        """Convert the value to a real python object"""
        return value

    def to_object(self, value, context=None):
        """Convert the value to python object, make validation possible"""
        return json.loads(value)

    def __eq__(self, other):
        if isinstance(other, dict):
            return self.get_schema() == other
        if isinstance(other, Schema):
            return self.get_schema() == other.get_schema()
        return id(self) == id(other)

    def invalidate(self, context=None):
        if hasattr(self, "_validator"):
            self._validator = None


class DeclarativeMetaclass(type):
    def __new__(mcls, name, bases, attrs):
        mcls.update_attrs(attrs, "_attrs", "Attrs")
        mcls.update_attrs(attrs, "_definitions", "Definitions")
        mcls.update_fields(attrs, '_fields', Schema)
        cls = super(DeclarativeMetaclass, mcls).__new__(
            mcls, name, bases, attrs
        )
        cls._attrs = mcls.get_inherited(cls, "_attrs", Schema)
        cls._definitions = mcls.get_inherited(cls, "_definitions", Schema)
        cls._fields = mcls.get_inherited(cls, "_fields", Schema)
        return cls

    @classmethod
    def update_attrs(mcls, attrs, name, clsname):
        Cls = attrs.pop(clsname, None)
        fields = lib.get_public_attributes(Cls)
        if name in attrs and attrs[name] is not None:
            attrs[name].update(fields)
        else:
            attrs[name] = fields

    @classmethod
    def update_fields(mcls, attrs, name, base):
        current_fields = []
        for key, value in list(attrs.items()):
            if isinstance(value, base):
                current_fields.append((key, value))
                attrs.pop(key)
        current_fields.sort(key=lambda x: x[1]._creation_index)
        if name in attrs and attrs[name] is not None:
            attrs[name].update(
                collections.OrderedDict(current_fields)
            )
        else:
            attrs[name] = collections.OrderedDict(current_fields) or None
        pass

    @classmethod
    def get_inherited(mcls, cls, name, base, remove_if_none=False):
        fields = collections.OrderedDict()
        mro = [c for c in cls.__mro__ if issubclass(c, base)]
        for basecls in reversed(mro):
            if hasattr(basecls, name) and getattr(basecls, name) is not None:
                fields.update(getattr(basecls, name))
            if remove_if_none:
                for attr, value in basecls.__dict__.items():
                    if value is None and attr in fields:
                        fields.pop(attr)
        return fields or None


@six.add_metaclass(DeclarativeMetaclass)
class Base(Schema):
    _type = None
    _attrs = None
    _definitions = None
    _properties = None

    def __init__(self, **attrs):
        super(Base, self).__init__(**attrs)

    def get_schema(self, context=None):
        if context:
            self.make_schema(context=context)
        if getattr(self, "_schema", None):
            return self._schema
        self._schema = self.make_schema(context=context)
        return self._schema

    def make_schema(self, context=None):
        if context is None:
            context = {}
        attr_exclude_tags = lib.ensure_set(self.get('exclude_tags'))
        ctx_exclude_tags = lib.ensure_set(context.get('exclude_tags'))
        exclude_tags = attr_exclude_tags | ctx_exclude_tags
        if attr_exclude_tags:
            context = context.copy()
            context['exclude_tags'] = exclude_tags

        schema = {"type": self._type}
        if self.get("null"):
            schema["type"] = [self._type, "null"]
        if self.get("enum"):
            schema["enum"] = self["enum"]
        if self.get("format"):
            schema["format"] = self["format"]
        if self.get("title"):
            schema["title"] = self["title"]
        if self.get("description"):
            schema["description"] = self["description"]
        if self.get("default"):
            schema["default"] = self["default"]
        if self._definitions:
            definitions = collections.OrderedDict()
            for name, prop in self._definitions.items():
                definitions[prop.get("name", name)] = \
                    prop.get_schema(context=context)
            schema["definitions"] = definitions
        if self._fields is not None:
            required = []
            properties = collections.OrderedDict()
            for key, prop in self._fields.items():
                if key in self.get('exclude', []):
                    continue
                if self.get('include'):
                    if key not in lib.ensure_list(self.get('include')):
                        continue
                if exclude_tags and prop.has_tags(exclude_tags):
                    continue
                name = prop.get("name", key)
                properties[name] = prop.get_schema(context=context)
                if prop.get('required'):
                    required.append(name)
            schema["properties"] = properties
            if required:
                schema['required'] = sorted(required)
        return schema

    def get_name(self, default=None):
        return self.get('name', default)

    def load(self, value, context=None):
        if isinstance(value, dict):
            by_name = {}
            for field, prop in self._fields.items():
                by_name[prop.get('name', field)] = prop
            for field in list(set(value) & set(by_name)):
                value[field] = by_name[field].to_object(
                    value[field], context=context
                )
            self.validate_json(value, context=context)
            self._value = self.to_python(value, context=context)
            return self._value
        return super(Base, self).load(value, context=context)

    def to_python(self, value, context=None):
        """Convert the value to a real python object"""
        if self._fields is not None:
            res = {}
            for field, schema in self._fields.items():
                name = schema.get('name', field)
                if name in value:
                    res[field] = schema.to_python(
                        value.pop(name), context=context
                    )
            res.update(value)
            return res
        return value

    def to_json(self, value, context=None):
        """Convert the value to a JSON compatible value"""
        if value is None:
            return None
        if self._fields is not None:
            res = {}
            value = value.copy()
            for field in list(set(value) & set(self._fields)):
                schema = self._fields.get(field)
                res[schema.get('name', field)] = \
                    schema.to_json(value.pop(field), context=context)
            res.update(value)
            return res
        return value

    def invalidate(self, context=None):
        super(Base, self).invalidate(context=None)
        if hasattr(self, "_schema"):
            self._schema = None


def _types_msg(instance, types, hint=''):
    reprs = []
    for type in types:
        try:
            reprs.append(repr(type["name"]))
        except Exception:
            reprs.append(repr(type))
    return "%r is not of type %s%s" % (instance, ", ".join(reprs), hint)


def _validate_type_draft4(validator, types, instance, schema):
    if isinstance(types, six.string_types):
        types = [types]
    if (
            'string' in types and
            'string' in schema.get('type') and
            schema.get('format') in [
                'date', 'datetime', 'time', 'duration', 'timestamp'
            ]
    ):
        if isinstance(instance, six.string_types):
            return
        if schema.get('format') == 'date' and \
                isinstance(instance, datetime.date):
            return
        if schema.get('format') == 'datetime' and \
                isinstance(instance, datetime.datetime):
            return
        if schema.get('format') == 'time' and \
                isinstance(instance, datetime.time):
            return
        if schema.get('format') == 'timestamp' and \
                isinstance(instance, (datetime.timedelta, int, float)):
            return
        if schema.get('format') == 'duration' and \
                isinstance(instance, (datetime.timedelta, int, float)):
            return

        json_format_name = schema.get('format')
        datetime_type_name = json_format_name.replace('-', '')
        hint = ' (for format %r strings, use a datetime.%s)' % (
            json_format_name, datetime_type_name
        )
        yield jsonschema.ValidationError(
            _types_msg(instance, types, hint)
        )

    if not any(validator.is_type(instance, type) for type in types):
        yield jsonschema.ValidationError(_types_msg(instance, types))


def _make_validator(schema):
    format_checker = jsonschema.FormatChecker(formats.draft4_format_checkers)
    validator_funcs = jsonschema.Draft4Validator.VALIDATORS
    validator_funcs[u'type'] = _validate_type_draft4
    meta_schema = jsonschema.Draft4Validator.META_SCHEMA
    validator_cls = jsonschema.validators.create(
        meta_schema=meta_schema,
        validators=validator_funcs,
        version="draft4",
    )
    validator_cls.check_schema(schema)
    return validator_cls(schema, format_checker=format_checker)



"""
This module introduce the base classes for reading and writing data based on
schema.
The preferred way is using reader is writer rather than using the
schema itself. It gives more flexibility and more extensibility.

.. todo::
    Note that this package in this version does nothing. Just give an early
    interface.
"""


class SchemaIO(object):
    """
    The schema IO gives chance to Schema remain independent from the
    serialisation method. Even the schema provide conversion still just
    based on primitive values.
    """

    def __init__(self, schema=None):
        self.schema = schema

    def get_schema(self, schema=None):
        """
        Gives back either the given schema or the schema which was defined
        in the constructor.
        If none of the specified it raises AttributeError.
        """
        if not (schema or self.schema):
            raise AttributeError('Schema not specified in either way')
        return schema or self.schema


class SchemaWriter(SchemaIO):
    """
    Writing schema
    .. code:: python

        sw = SchemaWriter(CustomSchema())
        encoded_data = sw.write({'custom': 'value'})
    """

    def write(self, data, schema=None):
        """
        With `self.get_schema()` select the proper schema, encode the given
        data then gives it back.
        """
        return self.get_schema(schema).dump(data)


class SchemaReader(SchemaIO):
    """
    Reading value based on schema
    .. code:: python

        sw = SchemaReader(CustomSchema())
        data = sw.write(<custom datastructure>)
    """

    def read(self, data, schema=None):
        """
        with `self.get_schema()` select the proper schema and read the data,
        validate the input and gives back the decoded value
        """
        return self.get_schema(schema).load(data)


class JSONWriter(SchemaWriter):
    pass


class JSONReader(SchemaReader):
    pass



import unittest

from .. import schemaio
from .. import types


class TestSchemaIO(unittest.TestCase):

    def test_init_store_schema(self):
        t = types.String()
        io = schemaio.SchemaIO(t)

        self.assertEqual(io.schema, t)

    def test_init_store_schema_none(self):
        io = schemaio.SchemaIO()

        self.assertEqual(io.schema, None)

    def test_get_schema_raise_error_with_none(self):
        io = schemaio.SchemaIO()

        with self.assertRaises(AttributeError):
            io.get_schema()

    def test_get_schema_gives_back_the_param(self):
        t = types.String()
        io = schemaio.SchemaIO()

        self.assertEqual(io.get_schema(t), t)

    def test_get_schema_gives_back_the_init_schema(self):
        t = types.String()
        io = schemaio.SchemaIO(t)

        self.assertEqual(io.get_schema(), t)

    def test_get_schema_prioritise_well(self):
        t1 = types.String()
        t2 = types.Integer()
        io = schemaio.SchemaIO(t1)

        self.assertEqual(io.get_schema(t2), t2)


class TestSchemaWriter(unittest.TestCase):

    def test_write(self):
        t1 = types.String()
        io = schemaio.SchemaWriter(t1)

        self.assertEqual(io.write('test'), '"test"')

    def test_write_with_given(self):
        t1 = types.String()
        io = schemaio.SchemaWriter()

        self.assertEqual(io.write('test', schema=t1), '"test"')


class TestSchemaReader(unittest.TestCase):

    def test_write(self):
        t1 = types.String()
        io = schemaio.SchemaReader(t1)

        self.assertEqual(io.read('"test"'), 'test')

    def test_write_with_given(self):
        t1 = types.String()
        io = schemaio.SchemaReader()

        self.assertEqual(io.read('"test"', schema=t1), 'test')


class TestJSONWriter(unittest.TestCase):

    def test_write(self):
        t1 = types.String()
        io = schemaio.JSONWriter(t1)

        self.assertEqual(io.write('test'), '"test"')

    def test_write_with_given(self):
        t1 = types.String()
        io = schemaio.JSONWriter()

        self.assertEqual(io.write('test', schema=t1), '"test"')


class TestJSONReader(unittest.TestCase):

    def test_write(self):
        t1 = types.String()
        io = schemaio.JSONReader(t1)

        self.assertEqual(io.read('"test"'), 'test')

    def test_write_with_given(self):
        t1 = types.String()
        io = schemaio.JSONReader()

        self.assertEqual(io.read('"test"', schema=t1), 'test')
